@@ -1,7 +1,7 @@
 /*
  *  herr.h
  *
- *  $Id: herr.h,v 1.11 2006/01/20 15:58:34 source Exp $
+ *  $Id: herr.h,v 1.12 2006/07/10 13:49:29 source Exp $
  *
  *  Error stack management functions
  *
@@ -90,6 +90,7 @@ typedef enum
     en_01S02,
     en_01S03,
     en_01S04,
+    en_01S08,
     en_07001,
     en_07006,
     en_08001,
@@ -123,6 +124,7 @@ typedef enum
     en_HY024,
     en_HY091,
     en_HY092,
+    en_HY110,
     en_HYC00,
     en_IM001,
     en_IM002,
