@@ -1,7 +1,7 @@
 /*
  *  odbc3.c
  *
- *  $Id: odbc3.c,v 1.23 2006/01/20 15:58:35 source Exp $
+ *  $Id: odbc3.c,v 1.30 2007/01/05 12:22:39 source Exp $
  *
  *  ODBC 3.x functions
  *
@@ -97,13 +97,19 @@
 #include <itrace.h>
 
 
+/*
+ * SQL_ATTR_CONNECTION_POOLING value
+ */
+SQLINTEGER _iodbcdm_attr_connection_pooling = SQL_CP_DEFAULT;
+
+
 /*
  *  Externals
  */
 SQLRETURN SQLAllocEnv_Internal (SQLHENV * phenv, int odbc_ver);
 SQLRETURN SQLFreeEnv_Internal (SQLHENV henv);
 SQLRETURN SQLAllocConnect_Internal (SQLHENV henv, SQLHDBC * phdbc);
-SQLRETURN SQLFreeConnect_Internal (SQLHDBC hdbc);
+SQLRETURN SQLFreeConnect_Internal (SQLHDBC hdbc, int ver);
 SQLRETURN SQLAllocStmt_Internal (SQLHDBC hdbc, SQLHSTMT * phstmt);
 SQLRETURN SQLFreeStmt_Internal (SQLHSTMT hstmt, SQLUSMALLINT fOption);
 SQLRETURN SQLTransact_Internal (SQLHENV henv, SQLHDBC hdbc, SQLUSMALLINT fType);
@@ -138,7 +144,17 @@ SQLAllocHandle_Internal (
       }
 
     case SQL_HANDLE_STMT:
-      return SQLAllocStmt_Internal (inputHandle, outputHandlePtr);
+      {
+	CONN (con, inputHandle);
+
+	if (!IS_VALID_HDBC (con))
+	  {
+	    return SQL_INVALID_HANDLE;
+	  }
+	CLEAR_ERRORS (con);
+
+        return SQLAllocStmt_Internal (inputHandle, outputHandlePtr);
+      }
 
     case SQL_HANDLE_DESC:
       {
@@ -147,6 +163,12 @@ SQLAllocHandle_Internal (
 	RETCODE retcode;
 	DESC_t *new_desc;
 
+	if (!IS_VALID_HDBC (con))
+	  {
+	    return SQL_INVALID_HANDLE;
+	  }
+	CLEAR_ERRORS (con);
+
 	if (((ENV_t *)(con->henv))->dodbc_ver == SQL_OV_ODBC2)
 	  {
 	    PUSHSQLERR (con->herr, en_HYC00);
@@ -173,10 +195,10 @@ SQLAllocHandle_Internal (
 	    return SQL_ERROR;
 	  }
 	memset (new_desc, 0, sizeof (DESC_t));
-	CALL_DRIVER (con, con, retcode, hproc, en_AllocHandle,
+	CALL_DRIVER (con, con, retcode, hproc,
 	    (handleType, con->dhdbc, &new_desc->dhdesc));
 
-	if (retcode != SQL_SUCCESS && retcode != SQL_SUCCESS_WITH_INFO)
+	if (!SQL_SUCCEEDED (retcode))
 	  {
 	    MEM_FREE (new_desc);
 	    return SQL_ERROR;
@@ -354,7 +376,7 @@ _SQLFreeHandle_DBC (
   ENTER_HDBC ((SQLHDBC) handle, 1,
       trace_SQLFreeHandle (TRACE_ENTER, handleType, handle));
 
-  retcode = SQLFreeConnect_Internal ((SQLHDBC) handle);
+  retcode = SQLFreeConnect_Internal ((SQLHDBC) handle, 3);
 
   LEAVE_HDBC ((SQLHDBC) handle, 1,
       trace_SQLFreeHandle (TRACE_LEAVE, handleType, handle);
@@ -426,7 +448,7 @@ _SQLFreeHandle_DESC (
 	  retcode = SQL_ERROR;
 	}
       else
-	CALL_DRIVER (pdbc, pdesc, retcode, hproc, en_FreeHandle,
+	CALL_DRIVER (pdbc, pdesc, retcode, hproc,
 	    (handleType, pdesc->dhdesc));
     }
 
@@ -500,7 +522,7 @@ SQLSetEnvAttr_Internal (SQLHENV environmentHandle,
 
   StringLength = StringLength; /*UNUSED*/
 
-  if (genv->hdbc)
+  if (genv != NULL && genv->hdbc)
     {
       PUSHSQLERR (genv->herr, en_HY010);
       return SQL_ERROR;
@@ -514,10 +536,12 @@ SQLSetEnvAttr_Internal (SQLHENV environmentHandle,
 	case SQL_CP_OFF:
 	case SQL_CP_ONE_PER_DRIVER:
 	case SQL_CP_ONE_PER_HENV:
-	  return SQL_SUCCESS;	/* not implemented yet */
+          _iodbcdm_attr_connection_pooling = (SQLINTEGER) (SQLULEN) ValuePtr;
+	  return SQL_SUCCESS;
 
 	default:
-	  PUSHSQLERR (genv->herr, en_HY024);
+	  if (genv)
+	    PUSHSQLERR (genv->herr, en_HY024);
 	  return SQL_ERROR;
 	}
 
@@ -526,7 +550,8 @@ SQLSetEnvAttr_Internal (SQLHENV environmentHandle,
 	{
 	case SQL_CP_STRICT_MATCH:
 	case SQL_CP_RELAXED_MATCH:
-	  return SQL_SUCCESS;	/* not implemented yet */
+          genv->cp_match = (SQLINTEGER) (SQLULEN) ValuePtr;
+	  return SQL_SUCCESS;
 
 	default:
 	  PUSHSQLERR (genv->herr, en_HY024);
@@ -575,22 +600,49 @@ SQLSetEnvAttr (
   SQLPOINTER		  ValuePtr,
   SQLINTEGER		  StringLength)
 {
-  ENTER_HENV (EnvironmentHandle,
-    trace_SQLSetEnvAttr (TRACE_ENTER,
-    	EnvironmentHandle,
-	Attribute,
-	ValuePtr, StringLength));
+  if (Attribute == SQL_ATTR_CONNECTION_POOLING)
+    {
+	SQLRETURN retcode = SQL_SUCCESS;
 
-  retcode = SQLSetEnvAttr_Internal (
-    	EnvironmentHandle,
-	Attribute,
-	ValuePtr, StringLength);
+	/* ENTER_HENV without EnvironmentHandle check */
+	ODBC_LOCK ();
+	TRACE (trace_SQLSetEnvAttr (TRACE_ENTER,
+	    EnvironmentHandle,
+	    Attribute,
+	    ValuePtr, StringLength));
 
-  LEAVE_HENV (EnvironmentHandle,
-    trace_SQLSetEnvAttr (TRACE_LEAVE,
-    	EnvironmentHandle,
-	Attribute,
-	ValuePtr, StringLength));
+        retcode = SQLSetEnvAttr_Internal (
+	      EnvironmentHandle,
+	      Attribute,
+	      ValuePtr, StringLength);
+
+	/* LEAVE_HENV without done: label */
+	TRACE (trace_SQLSetEnvAttr (TRACE_LEAVE,
+	    EnvironmentHandle,
+	    Attribute,
+	    ValuePtr, StringLength));
+	ODBC_UNLOCK ();
+	return retcode;
+    }
+  else
+    {
+      ENTER_HENV (EnvironmentHandle,
+        trace_SQLSetEnvAttr (TRACE_ENTER,
+	    EnvironmentHandle,
+	    Attribute,
+	    ValuePtr, StringLength));
+
+      retcode = SQLSetEnvAttr_Internal (
+	    EnvironmentHandle,
+	    Attribute,
+	    ValuePtr, StringLength);
+
+      LEAVE_HENV (EnvironmentHandle,
+        trace_SQLSetEnvAttr (TRACE_LEAVE,
+	    EnvironmentHandle,
+	    Attribute,
+	    ValuePtr, StringLength));
+   }
 }
 
 
@@ -625,13 +677,13 @@ SQLGetEnvAttr_Internal (
   if (Attribute == SQL_ATTR_CONNECTION_POOLING)
     {
       if (ValuePtr)
-	*((SQLUINTEGER *) ValuePtr) = SQL_CP_OFF;
+	*((SQLUINTEGER *) ValuePtr) = _iodbcdm_attr_connection_pooling;
       return SQL_SUCCESS;
     }
   if (Attribute == SQL_ATTR_CP_MATCH)
     {
       if (ValuePtr)
-	*((SQLUINTEGER *) ValuePtr) = SQL_CP_STRICT_MATCH;
+	*((SQLUINTEGER *) ValuePtr) = genv->cp_match;
       return SQL_SUCCESS;
     }
   if (Attribute == SQL_ATTR_OUTPUT_NTS)
@@ -655,7 +707,7 @@ SQLGetEnvAttr_Internal (
       if (hproc != SQL_NULL_HPROC)
 	{
 	  ENVR (env, con->henv);
-	  CALL_DRIVER (con, genv, retcode, hproc, en_GetEnvAttr,
+	  CALL_DRIVER (con, genv, retcode, hproc,
 	      (env->dhenv, Attribute, ValuePtr, BufferLength,
 		  StringLengthPtr));
 	  return retcode;
@@ -698,22 +750,49 @@ SQLGetEnvAttr (
   SQLINTEGER		  BufferLength,
   SQLINTEGER		* StringLengthPtr)
 {
-  ENTER_HENV (EnvironmentHandle,
-    trace_SQLGetEnvAttr (TRACE_ENTER,
-  	EnvironmentHandle,
-	Attribute,
-	ValuePtr, BufferLength, StringLengthPtr));
+  if (Attribute == SQL_ATTR_CONNECTION_POOLING)
+    {
+	SQLRETURN retcode = SQL_SUCCESS;
 
-  retcode = SQLGetEnvAttr_Internal (
-  	EnvironmentHandle,
-	Attribute,
-	ValuePtr, BufferLength, StringLengthPtr);
+	/* ENTER_HENV without EnvironmentHandle check */
+	ODBC_LOCK ();
+	TRACE (trace_SQLGetEnvAttr (TRACE_ENTER,
+	    EnvironmentHandle,
+	    Attribute,
+	    ValuePtr, BufferLength, StringLengthPtr));
 
-  LEAVE_HENV (EnvironmentHandle,
-    trace_SQLGetEnvAttr (TRACE_LEAVE,
-  	EnvironmentHandle,
-	Attribute,
-	ValuePtr, BufferLength, StringLengthPtr));
+        retcode = SQLGetEnvAttr_Internal (
+	      EnvironmentHandle,
+	      Attribute,
+	      ValuePtr, BufferLength, StringLengthPtr);
+
+	/* LEAVE_HENV without done: label */
+	TRACE (trace_SQLGetEnvAttr (TRACE_LEAVE,
+	    EnvironmentHandle,
+	    Attribute,
+	    ValuePtr, BufferLength, StringLengthPtr));
+	ODBC_UNLOCK ();
+	return retcode;
+    }
+  else
+    {
+      ENTER_HENV (EnvironmentHandle,
+        trace_SQLGetEnvAttr (TRACE_ENTER,
+  	    EnvironmentHandle,
+	    Attribute,
+	    ValuePtr, BufferLength, StringLengthPtr));
+
+      retcode = SQLGetEnvAttr_Internal (
+  	    EnvironmentHandle,
+	    Attribute,
+	    ValuePtr, BufferLength, StringLengthPtr);
+
+      LEAVE_HENV (EnvironmentHandle,
+        trace_SQLGetEnvAttr (TRACE_LEAVE,
+  	    EnvironmentHandle,
+	    Attribute,
+	    ValuePtr, BufferLength, StringLengthPtr));
+    }
 }
 
 
@@ -809,7 +888,6 @@ SQLGetStmtAttr_Internal (
       return SQL_SUCCESS;
 
     case SQL_ATTR_ROW_ARRAY_SIZE:
-
       if (((ENV_t *) ((DBC_t *) stmt->hdbc)->henv)->dodbc_ver == SQL_OV_ODBC3)
 	{
           CALL_UDRIVER(stmt->hdbc, stmt, retcode, hproc, 
@@ -1009,7 +1087,7 @@ SQLGetStmtAttr_Internal (
 	  if ((hproc = _iodbcdm_getproc (stmt->hdbc, en_GetStmtOption)) 
 	     != SQL_NULL_HPROC)
 	    {
-	      CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc, en_GetStmtOption,
+	      CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc,
 		  (stmt->dhstmt, Attribute, ValuePtr));
 	      return retcode;
 	    }
@@ -1017,7 +1095,7 @@ SQLGetStmtAttr_Internal (
 	  if ((hproc = _iodbcdm_getproc (stmt->hdbc, en_GetStmtOptionA)) 
 	     != SQL_NULL_HPROC)
 	    {
-	      CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc, en_GetStmtOptionA,
+	      CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc,
 		  (stmt->dhstmt, Attribute, ValuePtr));
 	      return retcode;
 	    }
@@ -1027,6 +1105,7 @@ SQLGetStmtAttr_Internal (
 	      return SQL_ERROR;
 	    }
 	}
+
     default:
       CALL_UDRIVER(stmt->hdbc, stmt, retcode, hproc, 
         penv->unicode_driver, en_GetStmtAttr, (stmt->dhstmt, Attribute, ValuePtr, 
@@ -1161,7 +1240,7 @@ SQLSetStmtAttr_Internal (
 	      PUSHSQLERR (stmt->herr, en_IM001);
 	      return SQL_ERROR;
             }
-	  if (retcode != SQL_SUCCESS || retcode != SQL_SUCCESS_WITH_INFO)
+	  if (!SQL_SUCCEEDED (retcode))
 	    return SQL_ERROR;
 
 	  stmt->desc[APP_PARAM_DESC] = SQL_NULL_HDESC;
@@ -1190,7 +1269,7 @@ SQLSetStmtAttr_Internal (
 	      PUSHSQLERR (stmt->herr, en_IM001);
 	      return SQL_ERROR;
             }
-	  if (retcode != SQL_SUCCESS || retcode != SQL_SUCCESS_WITH_INFO)
+	  if (!SQL_SUCCEEDED (retcode))
 	    return SQL_ERROR;
 
 	  stmt->desc[APP_PARAM_DESC] = (DESC_t *) ValuePtr;
@@ -1211,7 +1290,7 @@ SQLSetStmtAttr_Internal (
 	      PUSHSQLERR (stmt->herr, en_IM001);
 	      return SQL_ERROR;
             }
-	  if (retcode != SQL_SUCCESS || retcode != SQL_SUCCESS_WITH_INFO)
+	  if (!SQL_SUCCEEDED (retcode))
 	    return SQL_ERROR;
 
 	  stmt->desc[APP_ROW_DESC] = SQL_NULL_HDESC;
@@ -1240,7 +1319,7 @@ SQLSetStmtAttr_Internal (
 	      PUSHSQLERR (stmt->herr, en_IM001);
 	      return SQL_ERROR;
             }
-	  if (retcode != SQL_SUCCESS || retcode != SQL_SUCCESS_WITH_INFO)
+	  if (!SQL_SUCCEEDED (retcode))
 	    return SQL_ERROR;
 
 	  stmt->desc[APP_ROW_DESC] = (DESC_t *) ValuePtr;
@@ -1367,7 +1446,7 @@ SQLSetStmtAttr_Internal (
 	      return SQL_ERROR;
             }
 
-          if (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO)
+          if (SQL_SUCCEEDED (retcode))
             {
               stmt->rowset_size = Attribute;
               if (retcode == SQL_SUCCESS_WITH_INFO)
@@ -1409,10 +1488,10 @@ SQLSetStmtAttr_Internal (
 	  if ((hproc = _iodbcdm_getproc (stmt->hdbc, en_SetStmtOption))
 	      != SQL_NULL_HPROC)
 	    {
-	      CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc, en_SetStmtOption,
+	      CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc,
 		  (stmt->dhstmt, SQL_ROWSET_SIZE, stmt->row_array_size));
 
-              if (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO)
+              if (SQL_SUCCEEDED (retcode))
                 {
                   stmt->rowset_size = Attribute;;
                   if (retcode == SQL_SUCCESS_WITH_INFO)
@@ -1429,10 +1508,10 @@ SQLSetStmtAttr_Internal (
 	  if ((hproc = _iodbcdm_getproc (stmt->hdbc, en_SetStmtOptionA))
 	      != SQL_NULL_HPROC)
 	    {
-	      CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc, en_SetStmtOptionA,
+	      CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc,
 		  (stmt->dhstmt, SQL_ROWSET_SIZE, stmt->row_array_size));
 
-              if (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO)
+              if (SQL_SUCCEEDED (retcode))
                 {
                   stmt->rowset_size = Attribute;;
                   if (retcode == SQL_SUCCESS_WITH_INFO)
@@ -1507,8 +1586,7 @@ SQLSetStmtAttr_Internal (
 	      return SQL_ERROR;
             }
 
-          if (Attribute == SQL_ATTR_ROW_BIND_TYPE 
-              && (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO))
+          if (Attribute == SQL_ATTR_ROW_BIND_TYPE && SQL_SUCCEEDED (retcode))
             stmt->bind_type = Attribute;
 
           return retcode;
@@ -1518,10 +1596,9 @@ SQLSetStmtAttr_Internal (
 	  if ((hproc = _iodbcdm_getproc (stmt->hdbc, en_SetStmtOption))
 	      != SQL_NULL_HPROC)
 	    {
-	      CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc, en_SetStmtOption,
+	      CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc,
 		  (stmt->dhstmt, Attribute, ValuePtr));
-              if (Attribute == SQL_ATTR_ROW_BIND_TYPE 
-                  && (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO))
+              if (Attribute == SQL_ATTR_ROW_BIND_TYPE && SQL_SUCCEEDED (retcode))
                 stmt->bind_type = Attribute;
 	      return retcode;
 	    }
@@ -1529,10 +1606,9 @@ SQLSetStmtAttr_Internal (
 	  if ((hproc = _iodbcdm_getproc (stmt->hdbc, en_SetStmtOptionA))
 	      != SQL_NULL_HPROC)
 	    {
-	      CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc, en_SetStmtOptionA,
+	      CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc,
 		  (stmt->dhstmt, Attribute, ValuePtr));
-              if (Attribute == SQL_ATTR_ROW_BIND_TYPE 
-                  && (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO))
+              if (Attribute == SQL_ATTR_ROW_BIND_TYPE && SQL_SUCCEEDED (retcode))
                 stmt->bind_type = Attribute;
 	      return retcode;
 	    }
@@ -1546,7 +1622,7 @@ SQLSetStmtAttr_Internal (
       if ((hproc = _iodbcdm_getproc (stmt->hdbc, en_SetStmtOption))
          != SQL_NULL_HPROC)
 	{
-	  CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc, en_SetStmtOption,
+	  CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc,
 	      (stmt->dhstmt, Attribute, ValuePtr));
 	  return retcode;
 	}
@@ -1554,7 +1630,7 @@ SQLSetStmtAttr_Internal (
       if ((hproc = _iodbcdm_getproc (stmt->hdbc, en_SetStmtOptionA))
          != SQL_NULL_HPROC)
 	{
-	  CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc, en_SetStmtOptionA,
+	  CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc,
 	      (stmt->dhstmt, Attribute, ValuePtr));
 
 	  return retcode;
@@ -1672,9 +1748,12 @@ SQLSetConnectAttr_Internal (
   CONN (con, connectionHandle);
   ENVR (penv, con->henv);
   HPROC hproc = SQL_NULL_HPROC;
+  HPROC hproc2 = SQL_NULL_HPROC;
   SQLRETURN retcode = SQL_SUCCESS;
   void * _ValuePtr = NULL;
   SWORD unicode_driver = (penv ? penv->unicode_driver : 0);
+  SQLUINTEGER odbc_ver = ((GENV_t *) con->genv)->odbc_ver;
+  SQLUINTEGER dodbc_ver = ((ENV_t *) con->henv)->dodbc_ver;
 
   if (con->state == en_dbc_needdata)
     {
@@ -1709,11 +1788,16 @@ SQLSetConnectAttr_Internal (
         }
     }
 
-  CALL_UDRIVER(con, con, retcode, hproc, unicode_driver, 
-    en_SetConnectAttr, (con->dhdbc, Attribute, ValuePtr, StringLength));
-  if (hproc != SQL_NULL_HPROC)
+  GET_UHPROC(con, hproc2, en_SetConnectOption, unicode_driver);
+
+  if (dodbc_ver == SQL_OV_ODBC3 &&
+      (  odbc_ver == SQL_OV_ODBC3 
+       || (odbc_ver == SQL_OV_ODBC2 && hproc2 == SQL_NULL_HPROC)))
     {
-      return retcode;
+      CALL_UDRIVER(con, con, retcode, hproc, unicode_driver, 
+        en_SetConnectAttr, (con->dhdbc, Attribute, ValuePtr, StringLength));
+      if (hproc != SQL_NULL_HPROC)
+        return retcode;
     }
 
   switch (Attribute)
@@ -1824,10 +1908,13 @@ SQLGetConnectAttr_Internal (
   CONN (con, connectionHandle);
   ENVR (penv, con->henv);
   HPROC hproc = SQL_NULL_HPROC;
+  HPROC hproc2 = SQL_NULL_HPROC;
   RETCODE retcode = SQL_SUCCESS;
   void * _Value = NULL;
   void * valueOut = ValuePtr;
   SWORD unicode_driver = (penv ? penv->unicode_driver : 0);
+  SQLUINTEGER odbc_ver = ((GENV_t *) con->genv)->odbc_ver;
+  SQLUINTEGER dodbc_ver = ((ENV_t *) con->henv)->dodbc_ver;
 
   if (con->state == en_dbc_needdata)
     {
@@ -1870,15 +1957,23 @@ SQLGetConnectAttr_Internal (
         }
     }
 
-  CALL_UDRIVER(con, con, retcode, hproc, unicode_driver, 
-    en_GetConnectAttr, (con->dhdbc, Attribute, valueOut, StringLength,
-    StringLengthPtr));
+  GET_UHPROC(con, hproc2, en_GetConnectOption, unicode_driver);
+
+  if (dodbc_ver == SQL_OV_ODBC3 &&
+      (  odbc_ver == SQL_OV_ODBC3 
+       || (odbc_ver == SQL_OV_ODBC2 && hproc2 == SQL_NULL_HPROC)))
+    {
+      CALL_UDRIVER(con, con, retcode, hproc, unicode_driver, 
+        en_GetConnectAttr, (con->dhdbc, Attribute, valueOut, StringLength,
+        StringLengthPtr));
+    }
+
   if (hproc != SQL_NULL_HPROC)
     {
       if (ValuePtr 
-      && (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO)
-      &&  ((unicode_driver && waMode != 'W') 
-          || (!unicode_driver && waMode == 'W')))
+          && SQL_SUCCEEDED (retcode)
+          && ((unicode_driver && waMode != 'W') 
+              || (!unicode_driver && waMode == 'W')))
         {
           switch(Attribute)
             {
@@ -1915,7 +2010,7 @@ SQLGetConnectAttr_Internal (
   MEM_FREE(_Value);
 
   retcode = _iodbcdm_GetConnectOption (con, Attribute, ValuePtr, waMode);
-  if (retcode != SQL_SUCCESS || retcode != SQL_SUCCESS_WITH_INFO)
+  if (!SQL_SUCCEEDED (retcode))
     return retcode;
 
   if (StringLengthPtr)
@@ -2091,8 +2186,8 @@ SQLGetDescField_Internal (
     }    
 
   if (ValuePtr 
-      && (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO)
-      &&  ((penv->unicode_driver && waMode != 'W') 
+      && SQL_SUCCEEDED (retcode)
+      && ((penv->unicode_driver && waMode != 'W') 
           || (!penv->unicode_driver && waMode == 'W')))
     {
       switch(FieldIdentifier)
@@ -2429,8 +2524,8 @@ SQLGetDescRec_Internal (
     }    
 
   if (Name 
-      && (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO)
-      &&  ((penv->unicode_driver && waMode != 'W') 
+      && SQL_SUCCEEDED (retcode)
+      && ((penv->unicode_driver && waMode != 'W') 
           || (!penv->unicode_driver && waMode == 'W')))
     {
       if (waMode != 'W')
@@ -2631,7 +2726,7 @@ SQLSetDescRec_Internal (
       return SQL_ERROR;
     }
 
-  CALL_DRIVER (desc->hdbc, desc, retcode, hproc, en_SetDescRec,
+  CALL_DRIVER (desc->hdbc, desc, retcode, hproc,
       (desc->dhdesc, RecNumber, Type, SubType, Length, Precision, Scale, 
        Data, StringLength, Indicator));
 
@@ -2684,7 +2779,7 @@ SQLCopyDesc_Internal (
       return SQL_ERROR;
     }
 
-  CALL_DRIVER (desc->hdbc, desc, retcode, hproc, en_CopyDesc,
+  CALL_DRIVER (desc->hdbc, desc, retcode, hproc,
       (desc->dhdesc, desc1->dhdesc));
 
   return retcode;
@@ -2728,9 +2823,12 @@ SQLColAttribute_Internal (
   ENVR (penv, pdbc->henv);
   GENV (genv, pdbc->genv);
   HPROC hproc = SQL_NULL_HPROC;
+  HPROC hproc2 = SQL_NULL_HPROC;
   SQLRETURN retcode = SQL_SUCCESS;
   void * charAttrOut = CharacterAttributePtr;
   void * _charAttr = NULL;
+  SQLUINTEGER odbc_ver = ((GENV_t *) pdbc->genv)->odbc_ver;
+  SQLUINTEGER dodbc_ver = ((ENV_t *) pdbc->henv)->dodbc_ver;
 
   if ((penv->unicode_driver && waMode != 'W') 
       || (!penv->unicode_driver && waMode == 'W'))
@@ -2775,9 +2873,16 @@ SQLColAttribute_Internal (
         }
     }
 
-  CALL_UDRIVER(stmt->hdbc, stmt, retcode, hproc, penv->unicode_driver, 
-    en_ColAttribute, (stmt->dhstmt, ColumnNumber, FieldIdentifier, charAttrOut,
-    BufferLength, StringLengthPtr, NumericAttributePtr));
+  GET_UHPROC(stmt->hdbc, hproc2, en_ColAttributes, penv->unicode_driver);
+
+  if (dodbc_ver == SQL_OV_ODBC3 &&
+      (  odbc_ver == SQL_OV_ODBC3 
+       || (odbc_ver == SQL_OV_ODBC2 && hproc2 == SQL_NULL_HPROC)))
+    {
+      CALL_UDRIVER(stmt->hdbc, stmt, retcode, hproc, penv->unicode_driver, 
+        en_ColAttribute, (stmt->dhstmt, ColumnNumber, FieldIdentifier, 
+        charAttrOut, BufferLength, StringLengthPtr, NumericAttributePtr));
+    }
 
   if (hproc != SQL_NULL_HPROC)
     {
@@ -2793,7 +2898,7 @@ SQLColAttribute_Internal (
       }
 
       if (CharacterAttributePtr 
-          && (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO)
+          && SQL_SUCCEEDED (retcode)
           &&  ((penv->unicode_driver && waMode != 'W') 
               || (!penv->unicode_driver && waMode == 'W')))
         {
@@ -2875,7 +2980,7 @@ SQLColAttribute_Internal (
 	      PUSHSQLERR (stmt->herr, en_IM001);
 	      return SQL_ERROR;
 	    }
-	  CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc, en_NumResultCols,
+	  CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc,
 	      (stmt->dhstmt, NumericAttributePtr));
 	  return retcode;
 
@@ -3001,8 +3106,8 @@ SQLColAttribute_Internal (
       }
 
       if (CharacterAttributePtr 
-          && (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO)
-          &&  ((penv->unicode_driver && waMode != 'W') 
+          && SQL_SUCCEEDED (retcode)
+          && ((penv->unicode_driver && waMode != 'W') 
               || (!penv->unicode_driver && waMode == 'W')))
         {
           switch(FieldIdentifier)
@@ -3216,11 +3321,11 @@ SQLBulkOperations_Internal (
   hproc = _iodbcdm_getproc (stmt->hdbc, en_BulkOperations);
   if (hproc)
     {
-      CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc, en_BulkOperations,
+      CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc,
 	  (stmt->dhstmt, Operation));
       
       if (Operation == SQL_FETCH_BY_BOOKMARK 
-          && (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO))
+          && SQL_SUCCEEDED (retcode))
         _iodbcdm_ConvBindData (stmt);
       return retcode;
     }
@@ -3261,8 +3366,12 @@ SQLFetchScroll_Internal (
   SQLLEN		  fetchOffset)
 {
   STMT (stmt, statementHandle);
-  HPROC hproc;
+  HPROC hproc = SQL_NULL_HPROC;
   RETCODE retcode;
+  CONN (pdbc, stmt->hdbc);
+  HPROC hproc2 = SQL_NULL_HPROC;
+  SQLUINTEGER odbc_ver = ((GENV_t *) pdbc->genv)->odbc_ver;
+  SQLUINTEGER dodbc_ver = ((ENV_t *) pdbc->henv)->dodbc_ver;
 
   /* check arguments */
   switch (fetchOrientation)
@@ -3305,13 +3414,21 @@ SQLFetchScroll_Internal (
       return SQL_ERROR;
     }
 
-  hproc = _iodbcdm_getproc (stmt->hdbc, en_FetchScroll);
-  if (hproc)
+  hproc2 = _iodbcdm_getproc (stmt->hdbc, en_ExtendedFetch);
+
+  if (dodbc_ver == SQL_OV_ODBC3 &&
+      (  odbc_ver == SQL_OV_ODBC3 
+       || (odbc_ver == SQL_OV_ODBC2 && hproc2 == SQL_NULL_HPROC)))
     {
-      CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc, en_FetchScroll,
-	  (stmt->dhstmt, fetchOrientation, fetchOffset));
+      hproc = _iodbcdm_getproc (stmt->hdbc, en_FetchScroll);
+      if (hproc)
+        {
+          CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc,
+	      (stmt->dhstmt, fetchOrientation, fetchOffset));
+        }
     }
-  else
+
+  if (hproc == SQL_NULL_HPROC)
     {
       if (!stmt->row_status_ptr)
 	{
@@ -3403,7 +3520,7 @@ SQLFetchScroll (
 	FetchOrientation, 
 	FetchOffset);
 
-  if (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO)
+  if (SQL_SUCCEEDED (retcode))
     _iodbcdm_ConvBindData ((STMT_t *) StatementHandle);
 
   LEAVE_STMT (StatementHandle,
@@ -3433,11 +3550,12 @@ static SQLRETURN
 SQLCloseCursor_Internal (SQLHSTMT hstmt)
 {
   STMT (pstmt, hstmt);
-  CONN (pdbc, NULL);
+  CONN (pdbc, pstmt->hdbc);
   HPROC hproc = SQL_NULL_HPROC;
   SQLRETURN retcode = SQL_SUCCESS;
-
-  pdbc = (DBC_t *) (pstmt->hdbc);
+  HPROC hproc2 = SQL_NULL_HPROC;
+  SQLUINTEGER odbc_ver = ((GENV_t *) pdbc->genv)->odbc_ver;
+  SQLUINTEGER dodbc_ver = ((ENV_t *) pdbc->henv)->dodbc_ver;
 
   /* check state */
   if (pstmt->state >= en_stmt_needdata || pstmt->asyn_on != en_NullProc)
@@ -3447,25 +3565,31 @@ SQLCloseCursor_Internal (SQLHSTMT hstmt)
       return SQL_ERROR;
     }
 
-  hproc = _iodbcdm_getproc (pstmt->hdbc, en_CloseCursor);
+  hproc2 = _iodbcdm_getproc (pstmt->hdbc, en_FreeStmt);
 
-  if (hproc)
+  if (dodbc_ver == SQL_OV_ODBC3 &&
+      (  odbc_ver == SQL_OV_ODBC3 
+       || (odbc_ver == SQL_OV_ODBC2 && hproc2 == SQL_NULL_HPROC)))
     {
-      CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, en_CloseCursor,
-	  (pstmt->dhstmt));
+      hproc = _iodbcdm_getproc (pstmt->hdbc, en_CloseCursor);
+      if (hproc)
+        {
+          CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc,
+	      (pstmt->dhstmt));
+        }
     }
-  else
+
+  if (hproc == SQL_NULL_HPROC)
     {
       hproc = _iodbcdm_getproc (pstmt->hdbc, en_FreeStmt);
 
       if (hproc == SQL_NULL_HPROC)
 	{
 	  PUSHSQLERR (pstmt->herr, en_IM001);
-
 	  return SQL_ERROR;
 	}
 
-      CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, en_FreeStmt,
+      CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc,
 	  (pstmt->dhstmt, SQL_CLOSE));
     }
 
