@@ -1,7 +1,7 @@
 /*
  *  result.c
  *
- *  $Id: result.c,v 1.23 2006/01/20 15:58:35 source Exp $
+ *  $Id: result.c,v 1.28 2007/01/05 12:22:39 source Exp $
  *
  *  Prepare for getting query result
  *
@@ -190,16 +190,16 @@ SQLBindCol_Internal (
 
   if (icol != 0 && !penv->unicode_driver && nCType == SQL_C_WCHAR)
     {
-      CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, en_BindCol,
+      CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc,
           (pstmt->dhstmt, icol, SQL_C_CHAR, rgbValue, 
            cbValueMax, pcbValue));
     }
   else
-    CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, en_BindCol,
+    CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc,
         (pstmt->dhstmt, icol, nCType, rgbValue, cbValueMax, pcbValue));
 
   if (icol != 0 && !penv->unicode_driver && nCType == SQL_C_WCHAR 
-      && (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO))
+      && SQL_SUCCEEDED (retcode))
     {
       BIND_t tbind;
 
@@ -324,8 +324,8 @@ SQLGetCursorName_Internal (
     }
 
   if (szCursor 
-      && (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO)
-      &&  ((penv->unicode_driver && waMode != 'W') 
+      && SQL_SUCCEEDED (retcode)
+      && ((penv->unicode_driver && waMode != 'W') 
           || (!penv->unicode_driver && waMode == 'W')))
     {
       if (waMode != 'W')
@@ -448,7 +448,7 @@ SQLRowCount_Internal (
       return SQL_ERROR;
     }
 
-  CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, en_RowCount,
+  CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc,
       (pstmt->dhstmt, pcrow));
 
   return retcode;
@@ -508,7 +508,7 @@ _iodbcdm_NumResultCols (
       return SQL_ERROR;
     }
 
-  CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, en_NumResultCols,
+  CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc,
       (pstmt->dhstmt, &ccol));
 
   /* state transition */
@@ -726,8 +726,8 @@ SQLDescribeCol_Internal (
     *pfSqlType = _iodbcdm_map_sql_type (*pfSqlType, genv->odbc_ver);
 
   if (szColName 
-      && (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO)
-      &&  ((penv->unicode_driver && waMode != 'W') 
+      && SQL_SUCCEEDED (retcode)
+      && ((penv->unicode_driver && waMode != 'W') 
           || (!penv->unicode_driver && waMode == 'W')))
     {
       if (waMode != 'W')
@@ -927,12 +927,15 @@ SQLColAttributes_Internal (
   STMT(pstmt, hstmt);
   CONN (pdbc, pstmt->hdbc);
   ENVR (penv, pdbc->henv);
-  HPROC hproc = SQL_NULL_HPROC;
+  HPROC hproc2 = SQL_NULL_HPROC;
+  HPROC hproc3 = SQL_NULL_HPROC;
   SQLRETURN retcode = SQL_SUCCESS;
   void * _Desc = NULL;
   void * descOut = rgbDesc;
   sqlstcode_t sqlstat = en_00000;
   SQLUSMALLINT new_attr = fDescType;
+  SQLUINTEGER odbc_ver = ((GENV_t *) pdbc->genv)->odbc_ver;
+  SQLUINTEGER dodbc_ver = ((ENV_t *) pdbc->henv)->dodbc_ver;
 
   /* check arguments */
   if (icol == 0 && fDescType != SQL_COLUMN_COUNT)
@@ -1029,13 +1032,20 @@ SQLColAttributes_Internal (
 
   if (penv->unicode_driver)
     {
-      /* SQL_XXX_W */
+      hproc2 = _iodbcdm_getproc (pdbc, en_ColAttributesW);
 #if (ODBCVER >= 0x0300)
-      if ((hproc = _iodbcdm_getproc (pdbc, en_ColAttributeW)) 
-          != SQL_NULL_HPROC)
+      hproc3 = _iodbcdm_getproc (pdbc, en_ColAttributeW);
+#endif
+
+      if (odbc_ver == SQL_OV_ODBC2 && 
+          (  dodbc_ver == SQL_OV_ODBC2
+           || (dodbc_ver == SQL_OV_ODBC3 && hproc2 != SQL_NULL_HPROC)))
+        hproc3 = SQL_NULL_HPROC;
+
+#if (ODBCVER >= 0x0300)
+      if (hproc3 != SQL_NULL_HPROC)
         {
-          CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, 
-               en_ColAttributeW, (
+          CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc3, (
        	          pstmt->dhstmt, 
        	          icol, 
        	          new_attr, 
@@ -1044,13 +1054,17 @@ SQLColAttributes_Internal (
        	          pcbDesc, 
        	          pfDesc));
         }
-      else 
+      else
 #endif
-      if ((hproc = _iodbcdm_getproc (pdbc, en_ColAttributesW)) 
-          != SQL_NULL_HPROC)
         {
-          CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, 
-               en_ColAttributesW, (
+          if (hproc2 == SQL_NULL_HPROC)
+            {
+              _iodbcdm_FreeStmtParams(pstmt);
+              PUSHSQLERR (pstmt->herr, en_IM001);
+              return SQL_ERROR;
+            }
+
+          CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc2, (
       	          pstmt->dhstmt, 
       	          icol, 
       	          fDescType, 
@@ -1058,18 +1072,30 @@ SQLColAttributes_Internal (
       	          cbDescMax, 
       	          pcbDesc, 
       	          pfDesc));
-       	}
+        }
     }
   else
     {
       /* SQL_XXX */
       /* SQL_XXX_A */
+      hproc2 = _iodbcdm_getproc (pdbc, en_ColAttributes);
+      if (hproc2 == SQL_NULL_HPROC)
+        hproc2 = _iodbcdm_getproc (pdbc, en_ColAttributesA);
 #if (ODBCVER >= 0x0300)
-      if ((hproc = _iodbcdm_getproc (pdbc, en_ColAttribute)) 
-          != SQL_NULL_HPROC)
+      hproc3 = _iodbcdm_getproc (pdbc, en_ColAttribute);
+      if (hproc3 == SQL_NULL_HPROC)
+        hproc3 = _iodbcdm_getproc (pdbc, en_ColAttributeA);
+#endif
+      
+      if (odbc_ver == SQL_OV_ODBC2 && 
+          (  dodbc_ver == SQL_OV_ODBC2
+           || (dodbc_ver == SQL_OV_ODBC3 && hproc2 != SQL_NULL_HPROC)))
+        hproc3 = SQL_NULL_HPROC;
+
+#if (ODBCVER >= 0x0300)
+      if (hproc3 != SQL_NULL_HPROC)
         {
-          CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, 
-               en_ColAttribute, (
+          CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc3, (
        	          pstmt->dhstmt, 
        	          icol, 
        	          new_attr, 
@@ -1078,43 +1104,17 @@ SQLColAttributes_Internal (
        	          pcbDesc, 
        	          pfDesc));
         }
-      else 
-#endif
-      if ((hproc = _iodbcdm_getproc (pdbc, en_ColAttributes)) 
-          != SQL_NULL_HPROC)
-        {
-          CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, 
-               en_ColAttributes, (
-      	          pstmt->dhstmt, 
-      	          icol, 
-      	          fDescType, 
-      	          descOut, 
-      	          cbDescMax, 
-      	          pcbDesc, 
-      	          pfDesc));
-        }
       else
-#if (ODBCVER >= 0x0300)
-      if ((hproc = _iodbcdm_getproc (pdbc, en_ColAttributeA)) 
-          != SQL_NULL_HPROC)
-        {
-          CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, 
-               en_ColAttributeA, (
-       	          pstmt->dhstmt, 
-       	          icol, 
-       	          new_attr, 
-       	          descOut, 
-       	          cbDescMax, 
-       	          pcbDesc, 
-       	          pfDesc));
-        }
-      else 
 #endif
-      if ((hproc = _iodbcdm_getproc (pdbc, en_ColAttributesA)) 
-          != SQL_NULL_HPROC)
         {
-          CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, 
-               en_ColAttributesA, (
+          if (hproc2 == SQL_NULL_HPROC)
+            {
+              _iodbcdm_FreeStmtParams(pstmt);
+              PUSHSQLERR (pstmt->herr, en_IM001);
+              return SQL_ERROR;
+            }
+
+          CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc2, (
       	          pstmt->dhstmt, 
       	          icol, 
       	          fDescType, 
@@ -1125,16 +1125,9 @@ SQLColAttributes_Internal (
         }
     }
 
-  if (hproc == SQL_NULL_HPROC)
-    {
-      _iodbcdm_FreeStmtParams(pstmt);
-      PUSHSQLERR (pstmt->herr, en_IM001);
-      return SQL_ERROR;
-    }
-
   if (rgbDesc 
-      && (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO)
-      &&  ((penv->unicode_driver && waMode != 'W') 
+      && SQL_SUCCEEDED (retcode)
+      && ((penv->unicode_driver && waMode != 'W') 
           || (!penv->unicode_driver && waMode == 'W')))
     {
       switch(fDescType)
