@@ -1,13 +1,13 @@
 /*
  *  GetData.c
  *
- *  $Id: GetData.c,v 1.5 2004/07/06 15:26:07 source Exp $
+ *  $Id: GetData.c,v 1.7 2006/01/20 15:58:35 source Exp $
  *
  *  SQLGetData trace functions
  *
  *  The iODBC driver manager.
  *
- *  Copyright (C) 1996-2003 by OpenLink Software <iodbc@openlinksw.com>
+ *  Copyright (C) 1996-2006 by OpenLink Software <iodbc@openlinksw.com>
  *  All Rights Reserved.
  *
  *  This software is released under the terms of either of the following
@@ -16,6 +16,10 @@
  *      - GNU Library General Public License (see LICENSE.LGPL)
  *      - The BSD License (see LICENSE.BSD).
  *
+ *  Note that the only valid version of the LGPL license as far as this
+ *  project is concerned is the original GNU Library General Public License
+ *  Version 2, dated June 1991.
+ *
  *  While not mandated by the BSD license, any patches you make to the
  *  iODBC source code may be contributed back into the iODBC project
  *  at your discretion. Contributions will benefit the Open Source and
@@ -28,8 +32,8 @@
  *  ============================================
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Library General Public
- *  License as published by the Free Software Foundation; either
- *  version 2 of the License, or (at your option) any later version.
+ *  License as published by the Free Software Foundation; only
+ *  Version 2 of the License dated June 1991.
  *
  *  This library is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -38,7 +42,7 @@
  *
  *  You should have received a copy of the GNU Library General Public
  *  License along with this library; if not, write to the Free
- *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  *
  *
  *  The BSD License
@@ -69,6 +73,7 @@
  *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
+
 #include "trace.h"
 
 
@@ -126,7 +131,9 @@ _trace_data (
       break;
 
     case SQL_C_DATE:
+#if ODBCVER >= 0x0300
     case SQL_C_TYPE_DATE:
+#endif
       {
 	DATE_STRUCT *d = (DATE_STRUCT *) rgbValue;
 	sprintf (buf, "%04d-%02d-%02d", d->year, d->month, d->day);
@@ -171,6 +178,7 @@ _trace_data (
       break;
 #endif
 
+#if ODBCVER >= 0x0300
     case SQL_C_INTERVAL_DAY:
       {
 	SQL_INTERVAL_STRUCT *i = (SQL_INTERVAL_STRUCT *) rgbValue;
@@ -298,6 +306,7 @@ _trace_data (
 	trace_emit_string ((SQLCHAR *) buf, SQL_NTS, 0);
       }
       break;
+#endif
 
     case SQL_C_LONG:
     case SQL_C_SLONG:
@@ -317,10 +326,13 @@ _trace_data (
       break;
 
 
+#if ODBCVER >= 0x0300
     case SQL_C_NUMERIC:
       /* NOT YET */
       break;
+#endif
 
+#if ODBCVER >= 0x0300
     case SQL_C_SBIGINT:
 #if defined (ODBCINT64)
       {
@@ -340,6 +352,7 @@ _trace_data (
       }
 #endif
       break;
+#endif
 
     case SQL_C_SHORT:
     case SQL_C_SSHORT:
@@ -359,7 +372,9 @@ _trace_data (
       break;
 
     case SQL_C_TIME:
+#if ODBCVER >= 0x0300
     case SQL_C_TYPE_TIME:
+#endif
       {
 	TIME_STRUCT *t = (TIME_STRUCT *) rgbValue;
 	sprintf (buf, "%02d:%02d:%02d", t->hour, t->minute, t->second);
@@ -368,7 +383,9 @@ _trace_data (
       break;
 
     case SQL_C_TIMESTAMP:
+#if ODBCVER >= 0x0300
     case SQL_C_TYPE_TIMESTAMP:
+#endif
       {
 	TIMESTAMP_STRUCT *t = (TIMESTAMP_STRUCT *) rgbValue;
 	sprintf (buf, "%04d-%02d-%02d %02d:%02d:%02d.%06ld",
