@@ -1,7 +1,7 @@
 /*
  *  trace.c
  *
- *  $Id: trace.c,v 1.18 2006/01/24 00:09:25 source Exp $
+ *  $Id: trace.c,v 1.20 2006/12/21 11:24:58 source Exp $
  *
  *  Trace functions
  *
@@ -112,6 +112,13 @@
 #endif
 
 
+/*
+ *  Limit the size of the tracefile, to avoid a core dump when the 
+ *  the RLIMIT_FSIZE is reached.
+ */
+#define	MAX_TRACEFILE_LEN	1000000000L	/* about 1GB */
+
+
 /*
  *  Global trace flag
  */
@@ -147,6 +154,30 @@ trace_get_filename (void)
 }
 
 
+static void
+trace_strftime_now (char *buf, size_t buflen, char *format)
+{
+  time_t now;
+  struct tm *timeNow;
+#ifdef HAVE_LOCALTIME_R
+  struct tm keeptime;
+#endif
+
+  tzset ();
+  time (&now);
+
+#ifdef HAVE_LOCALTIME_R
+  timeNow = localtime_r (&now, &keeptime);
+#else
+  timeNow = localtime (&now);
+#endif
+
+  strftime (buf, buflen, format, timeNow);
+}
+
+
+
+
 void
 trace_set_filename (char *fname)
 {
@@ -245,14 +276,7 @@ trace_set_filename (char *fname)
 	    case 't':
 	    case 'T':
 	      {
-		time_t now;
-		struct tm *timeNow;
-
-		tzset ();
-		time (&now);
-		timeNow = localtime (&now);
-
-		strftime (tmp, sizeof (tmp), "%Y%m%d-%H%M%S", timeNow);
+		trace_strftime_now (tmp, sizeof (tmp), "%Y%m%d-%H%M%S");
 		strcpy (&buf[buf_pos], tmp);
 		buf_pos += strlen (tmp);
 		break;
@@ -339,20 +363,14 @@ trace_start(void)
   else
     {
       char mesgBuf[200];
-      time_t now;
-      struct tm *timeNow;
 
       trace_emit ("** iODBC Trace file\n");
 
       /*
        *  Show start time
        */
-      tzset ();
-      time (&now);
-      timeNow = localtime (&now);
-      strftime (mesgBuf,
-	  sizeof (mesgBuf), "** Trace started on %a %b %d %H:%M:%S %Y",
-	  timeNow);
+      trace_strftime_now (mesgBuf, sizeof (mesgBuf),
+	  "** Trace started on %a %b %d %H:%M:%S %Y");
       trace_emit ("%s\n", mesgBuf);
 
       /*
@@ -422,13 +440,8 @@ trace_stop(void)
       /*
        * Show end time
        */
-      tzset ();
-      time (&now);
-      timeNow = localtime (&now);
-      strftime (mesgBuf,
-	  sizeof (mesgBuf), "** Trace finished on %a %b %d %H:%M:%S %Y",
-	  timeNow);
-
+      trace_strftime_now (mesgBuf, sizeof (mesgBuf),
+	  "** Trace finished on %a %b %d %H:%M:%S %Y");
       trace_emit ("\n%s\n", mesgBuf);
 
       if (trace_fp_close)
@@ -657,6 +670,16 @@ _trace_print_function (int func, int trace_leave, int retcode)
   extern char *odbcapi_symtab[];
   char *ptr = "invalid retcode";
 
+  /*
+   * Guard against tracefile getting too big
+   */
+  if (trace_fp && ftell (trace_fp) > MAX_TRACEFILE_LEN)
+    {
+ 	trace_emit ("\n*** TRACEFILE LIMIT REACHED ***\n");
+	trace_stop ();
+ 	return;
+    }
+
   /*
    * Calculate timestamp
    */
